import os
import glob

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from audiomate.utils import textfile
from . import base
from . import downloader

# DOWNLOAD_URLS taken from https://voice.mozilla.org/de/datasets
DOWNLOAD_URLS = {
    'de': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/de.tar.gz',  # noqa
    'en': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/en.tar.gz',  # noqa
    'et': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/et.tar.gz',  # noqa
    'it': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/it.tar.gz',  # noqa
    'fr': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/fr.tar.gz',  # noqa
    'cy': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/cy.tar.gz',  # noqa
    'br': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/br.tar.gz',  # noqa
    'cv': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/cv.tar.gz',  # noqa
    'tr': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/tr.tar.gz',  # noqa
    'tt': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/tt.tar.gz',  # noqa
    'ky': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/ky.tar.gz',  # noqa
    'ga-IE': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/ga-IE.tar.gz',  # noqa
    'kab': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/kab.tar.gz',  # noqa
    'ca': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/ca.tar.gz',  # noqa
    'zh-TW': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/zh-TW.tar.gz',  # noqa
    'sl': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/sl.tar.gz',  # noqa
    'nl': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/nl.tar.gz',  # noqa
    'cnh': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/cnh.tar.gz',  # noqa
    'eo': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/eo.tar.gz',  # noqa
    'fa': 'https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4.s3.amazonaws.com/cv-corpus-4-2019-12-10/fa.tar.gz'  # noqa
}


class CommonVoiceDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the Common Voice Speech Dataset.

    Args:
      lang (string): Available languages are
        de|en|et|it|fr|cy|br|cv|tr|tt|ky|ga-IE|kab|ca|zh-TW|sl|nl|cnh|eo|fa
        See Common Voice dataset Version on download page for the
        corresponding language to the shortcut
    num_threads (int): Number of threads to use for download files.
    """

    def __init__(self, lang='de', num_threads=1):
        url = DOWNLOAD_URLS[lang]

        super(CommonVoiceDownloader, self).__init__(
            url,
            move_files_up=True,
            num_threads=num_threads
        )

    @classmethod
    def type(cls):
        return 'common-voice'


class CommonVoiceReader(base.CorpusReader):
    """
    Reader for the Common Voice Corpus.

    .. seealso::

       `Common-Voice <https://voice.mozilla.org/>`_
          Project Page
    """

    @classmethod
    def type(cls):
        return 'common-voice'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)
        subset_ids = CommonVoiceReader.get_subset_ids(path)

        for subset_idx in subset_ids:
            self.load_subset(corpus, path, subset_idx)

        return corpus

    @staticmethod
    def get_subset_ids(path):
        """ Return a list with ids of all available subsets (based on existing csv-files). """
        subset_ids = []

        for subset_path in glob.glob(os.path.join(path, '*.tsv')):
            file_name = os.path.split(subset_path)[1]
            basename = os.path.splitext(file_name)[0]

            # We don't want to include the invalidated files
            # since there maybe corrupt files
            if basename != 'invalidated':
                subset_ids.append(basename)

        return subset_ids

    def load_subset(self, corpus, path, subset_idx):
        """ Load subset into corpus. """
        csv_file = os.path.join(path, '{}.tsv'.format(subset_idx))
        subset_utt_ids = []

        entries = textfile.read_separated_lines_generator(
            csv_file,
            separator='\t',
            max_columns=8,
            ignore_lines_starting_with=['client_id'],
            keep_empty=True
        )

        for entry in entries:
            file_idx = self.create_assets_if_needed(
                corpus,
                path,
                entry
            )

            if file_idx is not None:
                subset_utt_ids.append(file_idx)

        utt_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utt_ids))
        subview = subset.Subview(corpus, filter_criteria=[utt_filter])
        corpus.import_subview(subset_idx, subview)

    def create_assets_if_needed(self, corpus, path, entry):
        """ Create File/Utterance/Issuer, if they not already exist and return utt-idx. """
        file_name = entry[1]
        file_idx, _ = os.path.splitext(file_name)

        if file_idx in self.invalid_utterance_ids:
            return None

        if file_idx not in corpus.utterances.keys():
            speaker_idx = entry[0]
            transcription = entry[2]

            if len(entry) >= 6:
                age = CommonVoiceReader.map_age(entry[5])
            else:
                age = issuers.AgeGroup.UNKNOWN

            if len(entry) >= 7:
                gender = CommonVoiceReader.map_gender(entry[6])
            else:
                gender = issuers.Gender.UNKNOWN

            file_path = os.path.join(path, 'clips', file_name)
            corpus.new_file(file_path, file_idx)

            if speaker_idx in corpus.issuers.keys():
                issuer = corpus.issuers[speaker_idx]
            else:
                issuer = issuers.Speaker(speaker_idx, gender=gender, age_group=age)
                corpus.import_issuers(issuer)

            utterance = corpus.new_utterance(file_idx, file_idx, issuer.idx)
            utterance.set_label_list(
                annotations.LabelList.create_single(
                    transcription,
                    idx=audiomate.corpus.LL_WORD_TRANSCRIPT
                )
            )

        return file_idx

    @staticmethod
    def map_age(age):
        """ Map age to correct age-group. """

        if age in [None, '']:
            return issuers.AgeGroup.UNKNOWN
        elif age == 'teens':
            return issuers.AgeGroup.YOUTH
        elif age in ['sixties', 'seventies', 'eighties', 'nineties']:
            return issuers.AgeGroup.SENIOR
        else:
            return issuers.AgeGroup.ADULT

    @staticmethod
    def map_gender(gender):
        """ Map gender to correct value. """

        if gender == 'male':
            return issuers.Gender.MALE
        elif gender == 'female':
            return issuers.Gender.FEMALE
        else:
            return issuers.Gender.UNKNOWN



import os
import shutil

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from audiomate.utils import download
from audiomate.utils import textfile

from . import base


SUBSETS = {
    'dev-clean': 'http://www.openslr.org/resources/12/dev-clean.tar.gz',
    'dev-other': 'http://www.openslr.org/resources/12/dev-other.tar.gz',
    'test-clean': 'http://www.openslr.org/resources/12/test-clean.tar.gz',
    'test-other': 'http://www.openslr.org/resources/12/test-other.tar.gz',
    'train-clean-100': 'http://www.openslr.org/resources/12/train-clean-100.tar.gz',
    'train-clean-360': 'http://www.openslr.org/resources/12/train-clean-360.tar.gz',
    'train-other-500': 'http://www.openslr.org/resources/12/train-other-500.tar.gz',
}


class LibriSpeechDownloader(base.CorpusDownloader):
    """
    Downloader for the LibriSpeech Dataset.

    Args:
        subsets (list): List of subsets to download.
                        If empty or ``None``, all subsets are downloaded.
        num_threads (int): Number of threads to use for download files.
        keep_archives (bool): If ``True``, keep downloaded archives after extraction.
    """

    def __init__(self, subsets=None, num_threads=1, keep_archives=False):
        self.subsets = subsets
        self.num_threads = num_threads
        self.keep_archives = keep_archives

    @classmethod
    def type(cls):
        return 'librispeech'

    def _download(self, target_path):
        if self.subsets is None or len(self.subsets) == 0:
            to_download = SUBSETS.keys()
        else:
            to_download = self.subsets

        os.makedirs(target_path, exist_ok=True)

        for subset_name in to_download:
            if subset_name in SUBSETS.keys():
                tmp_file = os.path.join(target_path, '{}.tar.gz'.format(subset_name))
                download.download_file(
                    SUBSETS[subset_name],
                    tmp_file,
                    num_threads=self.num_threads
                )
                download.extract_tar(tmp_file, target_path)
                extract_sub_path = os.path.join(target_path, 'LibriSpeech')

                for item in os.listdir(extract_sub_path):
                    item_path = os.path.join(extract_sub_path, item)
                    item_target_path = os.path.join(target_path, item)
                    shutil.move(item_path, item_target_path)

                shutil.rmtree('extract_sub_path', ignore_errors=True)

                if not self.keep_archives:
                    os.remove(tmp_file)


class LibriSpeechReader(base.CorpusReader):
    """
    Reader for the LibriSpeech Corpus.

    .. seealso::

       `LibriSpeech <https://www.openslr.org/12/>`_
          Project Page
    """

    @classmethod
    def type(cls):
        return 'librispeech'

    def _check_for_missing_files(self, path):
        files = []

        for name in ['SPEAKERS.TXT']:
            if not os.path.isfile(os.path.join(path, name)):
                files.append(name)

        return files

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        speaker_info_path = os.path.join(path, 'SPEAKERS.TXT')
        speakers = LibriSpeechReader.load_speakers(speaker_info_path)

        sf = LibriSpeechReader.available_subfolders

        for subset_idx, subset_path in sf(path, SUBSETS.keys()).items():
            subset_utt_ids = set()

            for speaker_idx, speaker_path in sf(subset_path).items():
                corpus.import_issuers(speakers[speaker_idx])

                for chapter_idx, chapter_path in sf(speaker_path).items():
                    transcript_path = os.path.join(
                        chapter_path,
                        '{}-{}.trans.txt'.format(speaker_idx, chapter_idx)
                    )
                    transcripts = LibriSpeechReader.load_transcripts(transcript_path)

                    for utt_idx, transcript in transcripts.items():
                        file_path = os.path.join(chapter_path, '{}.flac'.format(utt_idx))
                        corpus.new_file(file_path, utt_idx)

                        utterance = corpus.new_utterance(
                            utt_idx,
                            utt_idx,
                            speaker_idx
                        )

                        utterance.set_label_list(
                            annotations.LabelList.create_single(
                                transcript,
                                idx=audiomate.corpus.LL_WORD_TRANSCRIPT
                            )
                        )

                        subset_utt_ids.add(utt_idx)

            utt_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utt_ids))
            subview = subset.Subview(corpus, filter_criteria=[utt_filter])
            corpus.import_subview(subset_idx, subview)

        return corpus

    @staticmethod
    def available_subfolders(path, restrict=None):
        subfolders = {}

        for subfolder_name in os.listdir(path):
            if restrict is None or subfolder_name in restrict:
                subfolder_path = os.path.join(path, subfolder_name)

                if os.path.isdir(subfolder_path):
                    subfolders[subfolder_name] = subfolder_path

        return subfolders

    @staticmethod
    def load_speakers(path):
        entries = textfile.read_separated_lines_generator(
            path,
            separator='|',
            max_columns=5,
            ignore_lines_starting_with=[';']
        )

        speakers = {}

        for item in entries:
            idx = item[0].strip()
            gender_str = item[1].strip()

            if gender_str == 'M':
                gender = issuers.Gender.MALE
            elif gender_str == 'F':
                gender = issuers.Gender.FEMALE
            else:
                gender = issuers.Gender.UNKNOWN

            issuer = issuers.Speaker(
                idx,
                gender=gender
            )

            speakers[idx] = issuer

        return speakers

    @staticmethod
    def load_transcripts(path):
        entries = textfile.read_separated_lines_generator(
            path,
            separator=' ',
            max_columns=2
        )

        return {x[0].strip(): x[1].strip() for x in entries}



""" Logging functionaly to use within audiomate. """

import logging
import time


PROGRESS_LOGGER_DELAY_SECONDS = 5 * 60
PROGRESS_LOGGER_LEVEL = logging.INFO


class MateLogger(logging.getLoggerClass()):
    """
    Extension of the default Logger,
    that adds functionality to log progress of an iterable.
    """

    def progress(self, iterable, total=None, description=None):
        """
        Log the the progress on the loop over the given ``iterable``.

        Arguments:
            iterable (Iterable): The iterable to iterate over.
            total (int): Total number of elements in the iterable.
                         If ``None``, tries to get it via ``len(iterable)``.

        Returns:
            generator: Just pipes through the iterable as a generator which is given.
        """

        if not self.isEnabledFor(PROGRESS_LOGGER_LEVEL):
            for x in iterable:
                yield x

        else:
            if total is None and iterable is not None:
                try:
                    total = len(iterable)
                except (TypeError, AttributeError):
                    total = None

            last_log = time.time()
            item_count = 0

            if description is not None:
                self.log(PROGRESS_LOGGER_LEVEL, '[Start] %s', description)

            for index, x in enumerate(iterable):
                item_count += 1
                current_time = time.time()

                if current_time - last_log > PROGRESS_LOGGER_DELAY_SECONDS:
                    self.log(
                        PROGRESS_LOGGER_LEVEL,
                        '[%d / %d] %s', index + 1, total or '?', description or ''
                    )
                    last_log = current_time

                yield x

            self.log(
                PROGRESS_LOGGER_LEVEL,
                '[Done %d] %s', total or item_count, description or ''
            )


# Update Logger class with our customized one
logging.setLoggerClass(MateLogger)


def getLogger():
    """ Returns the library based logger.  """
    return logging.getLogger('audiomate')



import os

import pytest
from tests import resources

from audiomate.corpus.conversion import base


class DummyAudioFileConverter(base.AudioFileConverter):

    def __init__(self, a_ret, sampling_rate=16000, separate_file_per_utterance=False,
                 force_conversion=False):
        super(DummyAudioFileConverter, self).__init__(
            sampling_rate,
            separate_file_per_utterance,
            force_conversion
        )
        self.a_ret = a_ret
        self.a_log = []
        self.b_log = []

    def _file_extension(self):
        return 'wav'

    def _does_utt_match_target_format(self, utterance):
        self.a_log.append(utterance.idx)
        return self.a_ret[utterance.idx]

    def _convert_files(self, files):
        self.b_log.append(files)


@pytest.fixture
def ds():
    ds = resources.create_dataset()

    file_1_path = resources.sample_wav_file('wav_1.wav')
    file_2_path = resources.get_resource_path(('audio_formats', 'mp3_2_44_1k_16b.mp3'))
    file_3_path = resources.get_resource_path(('audio_formats', 'flac_1_16k_16b.flac'))
    file_4_path = resources.sample_wav_file('wav_4.wav')

    ds.tracks['wav-1'].path = file_1_path
    ds.tracks['wav_2'].path = file_2_path
    ds.tracks['wav_3'].path = file_3_path
    ds.tracks['wav_4'].path = file_4_path

    return ds


class TestAudioFileConverter:

    def test_convert(self, tmp_path, ds):
        c = DummyAudioFileConverter({
            'utt-1': True, 'utt-2': False, 'utt-3': False, 'utt-4': False, 'utt-5': True
        })
        res = c.convert(ds, str(tmp_path))

        assert sorted(c.a_log) == sorted(ds.utterances.keys())
        assert len(c.b_log) == 1

        assert sorted(c.b_log[0]) == sorted([
            (
                ds.tracks['wav_2'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav_2.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav_3.wav')
            )
        ])

        assert set(res.utterances.keys()) == set(ds.utterances.keys())
        assert set(res.tracks.keys()) == set(ds.tracks.keys())
        assert set(res.subviews.keys()) == set(ds.subviews.keys())

        assert res.utterances['utt-2'].track.path == os.path.join(str(tmp_path), 'wav_2.wav')

    def test_convert_separate_file_per_utterance(self, tmp_path, ds):
        c = DummyAudioFileConverter({
            'utt-1': True, 'utt-2': False, 'utt-3': False, 'utt-4': False, 'utt-5': True
        }, separate_file_per_utterance=True)
        res = c.convert(ds, str(tmp_path))

        assert sorted(c.a_log) == ['utt-1', 'utt-2', 'utt-5']
        assert len(c.b_log) == 1

        assert sorted(c.b_log[0]) == sorted([
            (
                ds.tracks['wav_2'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'utt-2.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                0.0,
                1.5,
                os.path.join(str(tmp_path), 'utt-3.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                1.5,
                2.5,
                os.path.join(str(tmp_path), 'utt-4.wav')
            ),
        ])

        assert set(res.utterances.keys()) == set(ds.utterances.keys())
        assert set(res.tracks.keys()) == {'wav-1', 'utt-2', 'utt-3', 'utt-4', 'wav_4'}
        assert set(res.subviews.keys()) == set(ds.subviews.keys())

        assert res.utterances['utt-2'].track.path == os.path.join(str(tmp_path), 'utt-2.wav')

    def test_convert_separate_file_per_utterance_and_force(self, tmp_path, ds):
        c = DummyAudioFileConverter({
            'utt-1': True, 'utt-2': False, 'utt-3': False, 'utt-4': False, 'utt-5': True
        }, separate_file_per_utterance=True, force_conversion=True)
        res = c.convert(ds, str(tmp_path))

        assert sorted(c.a_log) == []
        assert len(c.b_log) == 1

        assert sorted(c.b_log[0]) == sorted([
            (
                ds.tracks['wav-1'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'utt-1.wav')
            ),
            (
                ds.tracks['wav_2'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'utt-2.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                0.0,
                1.5,
                os.path.join(str(tmp_path), 'utt-3.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                1.5,
                2.5,
                os.path.join(str(tmp_path), 'utt-4.wav')
            ),
            (
                ds.tracks['wav_4'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'utt-5.wav')
            ),
        ])

        assert set(res.utterances.keys()) == set(ds.utterances.keys())
        assert set(res.tracks.keys()) == {'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'}
        assert set(res.subviews.keys()) == set(ds.subviews.keys())

        assert res.utterances['utt-2'].track.path == os.path.join(str(tmp_path), 'utt-2.wav')

    def test_convert_force(self, tmp_path, ds):
        c = DummyAudioFileConverter({
            'utt-1': True, 'utt-2': False, 'utt-3': False, 'utt-4': False, 'utt-5': True
        }, force_conversion=True)
        res = c.convert(ds, str(tmp_path))

        assert sorted(c.a_log) == []
        assert len(c.b_log) == 1

        assert sorted(c.b_log[0]) == sorted([
            (
                ds.tracks['wav-1'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav-1.wav')
            ),
            (
                ds.tracks['wav_2'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav_2.wav')
            ),
            (
                ds.tracks['wav_3'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav_3.wav')
            ),
            (
                ds.tracks['wav_4'].path,
                0,
                float('inf'),
                os.path.join(str(tmp_path), 'wav_4.wav')
            ),
        ])

        assert set(res.utterances.keys()) == set(ds.utterances.keys())
        assert set(res.tracks.keys()) == set(ds.tracks.keys())
        assert set(res.subviews.keys()) == set(ds.subviews.keys())

        assert res.utterances['utt-2'].track.path == os.path.join(str(tmp_path), 'wav_2.wav')



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io
from audiomate.corpus.io import librispeech

import requests_mock

from tests import resources
from . import reader_test as rt


class TestLibriSpeechDownloader:

    def test_download_all(self, tmpdir):
        target_folder = tmpdir.strpath
        downloader = io.LibriSpeechDownloader()

        with requests_mock.Mocker() as mock:
            # Return any size (doesn't matter, only for prints)
            for name, url in librispeech.SUBSETS.items():
                data_path = resources.get_resource_path([
                    'sample_archives',
                    'librispeech',
                    '{}.tar.gz'.format(name)
                ])
                with open(data_path, 'rb') as f:
                    data = f.read()
                mock.head(url, headers={'Content-Length': '100'})
                mock.get(url, content=data)

            downloader.download(target_folder)

        for name in librispeech.SUBSETS:
            assert os.path.isdir(os.path.join(target_folder, name))

        assert os.path.isfile(os.path.join(target_folder, 'BOOKS.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'CHAPTERS.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'LICENSE.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'README.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'SPEAKERS.TXT'))

    def test_download_two_subsets(self, tmpdir):
        target_folder = tmpdir.strpath
        downloader = io.LibriSpeechDownloader(subsets=['dev-clean', 'test-clean'])

        with requests_mock.Mocker() as mock:
            # Return any size (doesn't matter, only for prints)
            for name, url in librispeech.SUBSETS.items():
                data_path = resources.get_resource_path([
                    'sample_archives',
                    'librispeech',
                    '{}.tar.gz'.format(name)
                ])
                with open(data_path, 'rb') as f:
                    data = f.read()
                mock.head(url, headers={'Content-Length': '100'})
                mock.get(url, content=data)

            downloader.download(target_folder)

        for name in ['dev-clean', 'test-clean']:
            assert os.path.isdir(os.path.join(target_folder, name))

        assert os.path.isfile(os.path.join(target_folder, 'BOOKS.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'CHAPTERS.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'LICENSE.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'README.TXT'))
        assert os.path.isfile(os.path.join(target_folder, 'SPEAKERS.TXT'))


class TestLibriSpeechReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('librispeech')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH)

    EXPECTED_NUMBER_OF_TRACKS = 12
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('174-32-0001', 'dev-clean/174/32/174-32-0001.flac'),
        rt.ExpFileTrack('174-32-0002', 'dev-clean/174/32/174-32-0002.flac'),
        rt.ExpFileTrack('174-32-0003', 'dev-clean/174/32/174-32-0003.flac'),
        rt.ExpFileTrack('8842-12-0001', 'dev-clean/8842/12/8842-12-0001.flac'),
        rt.ExpFileTrack('8842-383-0001', 'dev-clean/8842/383/8842-383-0001.flac'),
        rt.ExpFileTrack('8842-383-0002', 'dev-clean/8842/383/8842-383-0002.flac'),
        rt.ExpFileTrack('14-32-0001', 'train-clean-360/14/32/14-32-0001.flac'),
        rt.ExpFileTrack('14-32-0002', 'train-clean-360/14/32/14-32-0002.flac'),
        rt.ExpFileTrack('14-384-0001', 'train-clean-360/14/384/14-384-0001.flac'),
        rt.ExpFileTrack('17-22-0001', 'train-clean-360/17/22/17-22-0001.flac'),
        rt.ExpFileTrack('17-22-0002', 'train-clean-360/17/22/17-22-0002.flac'),
        rt.ExpFileTrack('19-99-0001', 'train-clean-360/19/99/19-99-0001.flac'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 5
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('14', 3, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('17', 2, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('19', 1, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('8842', 3, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('174', 3, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 12
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('174-32-0001',   '174-32-0001',   '174',  0, float('inf')),
        rt.ExpUtterance('174-32-0002',   '174-32-0002',   '174',  0, float('inf')),
        rt.ExpUtterance('174-32-0003',   '174-32-0003',   '174',  0, float('inf')),
        rt.ExpUtterance('8842-12-0001',  '8842-12-0001',  '8842', 0, float('inf')),
        rt.ExpUtterance('8842-383-0001', '8842-383-0001', '8842', 0, float('inf')),
        rt.ExpUtterance('8842-383-0002', '8842-383-0002', '8842', 0, float('inf')),
        rt.ExpUtterance('14-32-0001',    '14-32-0001',    '14',   0, float('inf')),
        rt.ExpUtterance('14-32-0002',    '14-32-0002',    '14',   0, float('inf')),
        rt.ExpUtterance('14-384-0001',   '14-384-0001',   '14',   0, float('inf')),
        rt.ExpUtterance('17-22-0001',    '17-22-0001',    '17',   0, float('inf')),
        rt.ExpUtterance('17-22-0002',    '17-22-0002',    '17',   0, float('inf')),
        rt.ExpUtterance('19-99-0001',    '19-99-0001',    '19',   0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '174-32-0001': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '174-32-0002': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '174-32-0003': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '8842-12-0001': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)
        ],
        '14-384-0001': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)
        ],
    }

    EXPECTED_LABELS = {
        '174-32-0001': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'A', 0, float('inf')),
        ],
        '174-32-0002': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'B', 0, float('inf')),
        ],
        '174-32-0003': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'C', 0, float('inf')),
        ],
        '8842-12-0001': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'XY', 0, float('inf')),
        ],
        '14-384-0001': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'HOPE FOR', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 2

    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('dev-clean', [
            '174-32-0001',
            '174-32-0002',
            '174-32-0003',
            '8842-12-0001',
            '8842-383-0001',
            '8842-383-0002',
        ]),
        rt.ExpSubview('train-clean-360', [
            '14-32-0001',
            '14-32-0002',
            '14-384-0001',
            '17-22-0001',
            '17-22-0002',
            '19-99-0001',
        ]),
    ]

    def load(self):
        reader = librispeech.LibriSpeechReader()
        return reader.load(self.SAMPLE_PATH)



import os
import collections

import audiomate
from . import base
from audiomate.utils import textfile
from audiomate.corpus import conversion


class MozillaDeepSpeechWriter(base.CorpusWriter):
    """
    Writes files to use for training with Mozilla DeepSpeech (https://github.com/mozilla/DeepSpeech).

    Since it is expected that every utterance is in a separate file,
    any utterances that are not in separate file in the original corpus,
    are extracted into a separate file in the subfolder `audio` of the target path.

    Args:
        no_audio_check (bool): If ``True``, the audio is not checked for correct format.
        export_all_audio (bool): If ``True``, all utterances are exported,
                                 whether they are in a separate file
                                 already or not.
        transcription_label_list_idx (str): The transcriptions are used from the label-list with this id.
        sampling_rate (int): Target sampling rate to use.
        num_workers (int): Number of processes to use to process utterances.
    """

    def __init__(self, export_all_audio=False,
                 no_audio_check=False,
                 transcription_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT,
                 sampling_rate=16000, num_workers=1):
        self.no_audio_check = no_audio_check
        self.export_all_audio = export_all_audio
        self.transcription_label_list_idx = transcription_label_list_idx
        self.sampling_rate = sampling_rate
        self.num_workers = num_workers

        self.converter = conversion.WavAudioFileConverter(
            self.num_workers,
            self.sampling_rate,
            separate_file_per_utterance=True,
            force_conversion=self.export_all_audio
        )

    @classmethod
    def type(cls):
        return 'mozilla-deepspeech'

    def _save(self, corpus, path):
        target_audio_path = os.path.join(path, 'audio')
        os.makedirs(target_audio_path, exist_ok=True)

        # Convert all files
        if not self.no_audio_check:
            corpus = self.converter.convert(corpus, target_audio_path)

        records = []

        subset_utterance_ids = {idx: set(subset.utterances.keys()) for idx, subset in corpus.subviews.items()}
        subset_records = collections.defaultdict(list)

        for utterance_idx in sorted(corpus.utterances.keys()):
            utterance = corpus.utterances[utterance_idx]
            transcript = utterance.label_lists[self.transcription_label_list_idx].join()
            audio_path = utterance.track.path
            size = os.stat(audio_path).st_size

            record = [audio_path, size, transcript]
            records.append(record)

            # Check / Add to subview lists
            for subset_idx, utt_ids in subset_utterance_ids.items():
                if utterance_idx in utt_ids:
                    subset_records[subset_idx].append(record)

        # Write full list
        records.insert(0, ['wav_filename', 'wav_filesize', 'transcript'])
        records_path = os.path.join(path, 'all.csv')
        textfile.write_separated_lines(records_path, records, separator=',', sort_by_column=-1)

        # Write subset lists
        for subset_idx, records in subset_records.items():
            if len(records) > 0:
                records.insert(0, ['wav_filename', 'wav_filesize', 'transcript'])
                subset_file_path = os.path.join(path, '{}.csv'.format(subset_idx))
                textfile.write_separated_lines(subset_file_path, records, separator=',', sort_by_column=-1)



import os
import re
import xml.etree.ElementTree as ET

from . import base

import audiomate
from audiomate.utils import jsonfile
from audiomate import issuers
from audiomate import annotations
from audiomate import tracks
from . import downloader


URLS = {
    'de': 'https://www2.informatik.uni-hamburg.de/nats/pub/SWC/SWC_German.tar',
    'en': 'https://www2.informatik.uni-hamburg.de/nats/pub/SWC/SWC_English.tar',
    'nl': 'https://www2.informatik.uni-hamburg.de/nats/pub/SWC/SWC_Dutch.tar'
}

READER_NAME_PATTERN = re.compile(r'user_name\s+=\s+(.*?)\n')
READER_GENDER_PATTERN = re.compile(r'(gender|geschlecht)\s+=\s+(.*?)\n')

MIN_SEGMENT_DURATION = 1.0


class SWCDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the MUSAN Corpus.

    Args:
        lang (str): The language to download.
    """

    def __init__(self, lang='de'):
        url = URLS[lang]

        super(SWCDownloader, self).__init__(
            url,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'swc'


class SWCReader(base.CorpusReader):
    """ Reader for the Spoken Wikipedia Corpus. """

    @classmethod
    def type(cls):
        return 'swc'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus()

        article_paths = sorted(self.get_articles(path))
        reader_map = {}
        file_map = {}

        for article_path in article_paths:
            audio_files = self.get_audio_file_info(article_path)
            reader_name, reader_gender = self.get_reader_info(article_path)
            segments = self.get_segments(article_path)

            if reader_name not in reader_map.keys():
                speaker = issuers.Speaker(
                    '{:0>8}'.format(len(reader_map)),
                    gender=reader_gender
                )
                reader_map[reader_name] = speaker
                corpus.import_issuers(speaker)
            else:
                speaker = reader_map[reader_name]

            for start, end, text in segments:
                file_path = self.find_audio_file_for_segment(start, end, audio_files)

                if file_path is not None:
                    if file_path not in file_map.keys():
                        track = tracks.FileTrack(
                            '{:0>10}'.format(len(file_map)),
                            file_path
                        )
                        file_map[file_path] = track
                        corpus.import_tracks(track)
                    else:
                        track = file_map[file_path]

                    track_offset = audio_files[file_path]
                    utt_start = start - track_offset
                    utt_end = end - track_offset

                    utt_idx = '{}_{}_{}_{}'.format(
                        speaker.idx,
                        track.idx,
                        int(start * 1000),
                        int(end * 1000)
                    )

                    if utt_idx not in self.invalid_utterance_ids:
                        utt = corpus.new_utterance(
                            utt_idx,
                            track.idx,
                            issuer_idx=speaker.idx,
                            start=utt_start,
                            end=utt_end
                        )

                        ll = annotations.LabelList.create_single(
                            text,
                            audiomate.corpus.LL_WORD_TRANSCRIPT
                        )

                        utt.set_label_list(ll)

        return audiomate.Corpus.from_corpus(corpus)

    def get_articles(self, path):
        """ Return the list of article-paths """
        article_paths = []

        for dirname in os.listdir(path):
            dirpath = os.path.join(path, dirname)

            audio_meta_file = os.path.join(dirpath, 'audiometa.txt')
            info_file = os.path.join(dirpath, 'info.json')
            align_file = os.path.join(dirpath, 'aligned.swc')

            if os.path.isfile(audio_meta_file) and \
                    os.path.isfile(info_file) and \
                    os.path.isfile(align_file):
                article_paths.append(dirpath)

        return article_paths

    def get_audio_file_info(self, article_path):
        """
        Return info about the audio files.
        List of tuples with (path, offset).
        """

        info_path = os.path.join(article_path, 'info.json')
        info = jsonfile.read_json_file(info_path)
        audio_files = {}

        if len(info['audio_files']) == 1:
            path = os.path.join(article_path, 'audio.ogg')

            if 'offset' not in info['audio_files'][0].keys():
                return {}

            offset = info['audio_files'][0]['offset']
            audio_files[path] = offset

        else:
            for i, af in enumerate(info['audio_files']):
                path = os.path.join(article_path, 'audio{}.ogg'.format(i+1))
                offset = af['offset']
                audio_files[path] = offset

        return audio_files

    def get_reader_info(self, article_path):
        """ Return info about the reader of the article. """

        meta_path = os.path.join(article_path, 'audiometa.txt')
        with open(meta_path, 'r', encoding='utf-8') as f:
            content = f.read()

        name_match = READER_NAME_PATTERN.search(content)

        if name_match is not None:
            name = name_match.group(1).strip()
        else:
            name = os.path.basename(article_path)

        gender_match = READER_GENDER_PATTERN.search(content)
        gender_str = ''

        if gender_match is not None:
            gender_str = gender_match.group(2).strip()

        if gender_str.lower() in ['male', 'männlich', 'mänlich', 'mann', 'm', 'malee', 'männ', 'maennlich']:
            gender = issuers.Gender.MALE
        elif gender_str.lower() in ['female', 'weiblich']:
            gender = issuers.Gender.FEMALE
        else:
            gender = issuers.Gender.UNKNOWN

        return name, gender

    def find_audio_file_for_segment(self, start, end, audio_files):
        """
        Find the correct audio file for the segment.
        Return index of matching audio file.
        """

        items = sorted(audio_files.items(), key=lambda x: x[1])

        for i in range(len(items) - 1):
            if end <= items[i+1][1]:
                # Segment belongs to audiofile i
                if start >= items[i][1]:
                    return items[i][0]

                # Segment crosses audiofile boundaries, ignore
                else:
                    return None

        if start >= items[-1][1]:
            return items[-1][0]
        else:
            return None

    def get_segments(self, article_path):
        """
        Parse segments from alignment file.
        Return list with tuples (start, end, text).
        """
        aligned_path = os.path.join(article_path, 'aligned.swc')
        tree = ET.parse(aligned_path)
        root = tree.getroot()

        segments = []

        for p in root:
            if p.tag == 'd':
                q = self.parse_element(p)
                segments.extend(q)

        # Filter segments on minimal duration
        segments = [s for s in segments if s[1] - s[0] > MIN_SEGMENT_DURATION]

        return segments

    def parse_element(self, parent):
        segments = []

        for c in parent:
            if c.tag in ['extra', 'section', 'sectioncontent', 'p']:
                q = self.parse_element(c)
                segments.extend(q)
            elif c.tag in ['sectiontitle', 's']:
                q = self.parse_sentence(c)
                segments.extend(q)

        return segments

    def parse_sentence(self, element):
        tokens = []

        for index, token in enumerate(element):
            if token.tag == 't':
                tokens.append((index, token))

        return self.parse_tokens(tokens)

    def parse_tokens(self, tokens):
        """ Get segments from sequence of tokens. """
        valid_tokens = self.get_valid_tokens(tokens)

        if len(valid_tokens) <= 0:
            return []

        segmented_tokens = []
        current_tokens = [valid_tokens[0]]

        for t in valid_tokens[1:]:
            # Tokens have consecutive indices
            # so same segment
            if current_tokens[-1][0] + 1 == t[0]:
                current_tokens.append(t)

            # start new segment
            else:
                segmented_tokens.append(current_tokens)
                current_tokens = [t]

        segmented_tokens.append(current_tokens)

        segments = []

        for token_group in segmented_tokens:
            text = ' '.join(t[3].strip() for t in token_group)
            start = token_group[0][1]
            end = token_group[-1][2]
            segments.append((start, end, text))

        return segments

    def get_valid_tokens(self, tokens):
        """
        Return only valid tokens
        (have a normalization with start and end time).
        """
        valid_tokens = []

        for index, token in tokens:
            normalizations = self.parse_normalizations(token)

            if len(normalizations) > 0:
                normalized_text = ' '.join(n[0] for n in normalizations)
                start = normalizations[0][1]
                end = normalizations[-1][2]

                if start != -1 and end != -1:
                    valid_tokens.append((
                        index,
                        start,
                        end,
                        normalized_text
                    ))

        return valid_tokens

    def parse_normalizations(self, token):
        """
        Parse normalizations of a token.
        Return list of tuples (text, start, end).
        """
        norms = []

        for c in token:
            if c.tag == 'n':
                pronunciation = c.attrib['pronunciation']
                start = -1
                end = -1

                if 'start' in c.attrib.keys():
                    start = int(c.attrib['start']) / 1000.0

                if 'end' in c.attrib.keys():
                    end = int(c.attrib['end']) / 1000.0

                norms.append((
                    pronunciation,
                    start,
                    end
                ))

        return norms



import json
import os
import pytest

import audiomate
from audiomate.issuers import Speaker, Issuer, Gender
from audiomate.annotations import LabelList, Label
from audiomate.corpus.subset import subview
from audiomate.corpus import io

from tests import resources


def create_sample_dataset(temp_dir):
    ds = audiomate.Corpus(str(temp_dir))

    file_1_path = resources.sample_wav_file('wav_1.wav')
    file_2_path = resources.sample_wav_file('wav_2.wav')
    file_3_path = resources.get_resource_path(['audio_formats', 'flac_1_16k_16b.flac'])

    file_1 = ds.new_file(file_1_path, track_idx='wav_1')
    file_2 = ds.new_file(file_2_path, track_idx='wav_2')
    file_3 = ds.new_file(file_3_path, track_idx='wav_3')

    issuer_1 = Speaker('spk-1', gender=Gender.MALE)
    issuer_2 = Speaker('spk-2', gender=Gender.FEMALE)
    issuer_3 = Issuer('spk-3')

    ds.import_issuers([issuer_1, issuer_2, issuer_3])

    # 2.5951875
    utt_1 = ds.new_utterance('utt-1', file_1.idx, issuer_idx=issuer_1.idx)
    utt_2 = ds.new_utterance('utt-2', file_2.idx, issuer_idx=issuer_2.idx, start=0, end=1.5)
    utt_3 = ds.new_utterance('utt-3', file_2.idx, issuer_idx=issuer_2.idx, start=1.5, end=2.5)
    # 5.0416875
    utt_4 = ds.new_utterance('utt-4', file_3.idx, issuer_idx=issuer_3.idx)

    utt_1.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who am i')]))
    utt_2.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who are you')]))
    utt_3.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who is he')]))
    utt_4.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who are they')]))

    train_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-1', 'utt-2', 'utt-3'})
    sv_train = subview.Subview(ds, filter_criteria=[train_filter])

    dev_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-4'})
    sv_dev = subview.Subview(ds, filter_criteria=[dev_filter])

    ds.import_subview('train', sv_train)
    ds.import_subview('dev', sv_dev)

    return ds


class TestNvidiaJasperWriter:

    def test_default_params(self, tmp_path):
        writer = io.NvidiaJasperWriter()
        ds = create_sample_dataset(tmp_path)

        writer.save(ds, str(tmp_path))

        all_file = tmp_path / 'all.json'
        train_file = tmp_path / 'train.json'
        dev_file = tmp_path / 'dev.json'

        assert all_file.exists()
        assert train_file.exists()
        assert dev_file.exists()

        all_content = json.loads(all_file.read_text())
        assert len(all_content) == 4

        assert all_content[0] == {
            'transcript': 'who is he',
            'files': [{
                'fname': 'audio/utt-3.wav',
                'channels': 1,
                'sample_rate': 16000,
                'duration': pytest.approx(1.0),
                'num_samples': 16000,
                'speed': 1}],
            'original_duration': pytest.approx(1.0),
            'original_num_samples': 16000,
            'utt_idx': 'utt-3'
        }
        assert (tmp_path / all_content[0]['files'][0]['fname']).exists()

        assert all_content[1] == {
            'transcript': 'who are you',
            'files': [{
                'fname': 'audio/utt-2.wav',
                'channels': 1,
                'sample_rate': 16000,
                'duration': pytest.approx(1.5),
                'num_samples': 24000,
                'speed': 1
            }],
            'original_duration': pytest.approx(1.5),
            'original_num_samples': 24000,
            'utt_idx': 'utt-2'
        }
        assert (tmp_path / all_content[1]['files'][0]['fname']).exists()

        assert all_content[2] == {
            'transcript': 'who am i',
            'files': [{
                'fname': os.path.relpath(ds.tracks['wav_1'].path, str(tmp_path)),
                'channels': 1,
                'sample_rate': 16000,
                'duration': pytest.approx(2.5951875),
                'num_samples': 41523,
                'speed': 1
            }],
            'original_duration': pytest.approx(2.5951875),
            'original_num_samples': 41523,
            'utt_idx': 'utt-1'
        }
        abspath = os.path.abspath(os.path.join(str(tmp_path), all_content[2]['files'][0]['fname']))
        assert os.path.isfile(abspath)

        # Use large margin for duration/num_samples since some backends return different results
        assert all_content[3] == {
            'transcript': 'who are they',
            'files': [{
                'fname': 'audio/utt-4.wav',
                'channels': 1,
                'sample_rate': 16000,
                'duration': pytest.approx(6.464, abs=1e-1),
                'num_samples': pytest.approx(103424, abs=1600),
                'speed': 1
            }],
            'original_duration': pytest.approx(6.464, abs=1e-1),
            'original_num_samples': pytest.approx(103424, abs=1600),
            'utt_idx': 'utt-4'
        }
        assert (tmp_path / all_content[3]['files'][0]['fname']).exists()

        train_content = json.loads(train_file.read_text())
        assert len(train_content) == 3

        dev_content = json.loads(dev_file.read_text())
        assert len(dev_content) == 1

    def test_export_all(self, tmp_path):
        writer = io.NvidiaJasperWriter(export_all_audio=True)
        ds = create_sample_dataset(tmp_path)

        writer.save(ds, str(tmp_path))

        all_file = tmp_path / 'all.json'
        train_file = tmp_path / 'train.json'
        dev_file = tmp_path / 'dev.json'

        assert all_file.exists()
        assert train_file.exists()
        assert dev_file.exists()

        all_content = json.loads(all_file.read_text())
        assert len(all_content) == 4

        assert all_content[2] == {
            'transcript': 'who am i',
            'files': [{
                'fname': 'audio/utt-1.wav',
                'channels': 1,
                'sample_rate': 16000,
                'duration': pytest.approx(2.5951875),
                'num_samples': 41523,
                'speed': 1
            }],
            'original_duration': pytest.approx(2.5951875),
            'original_num_samples': 41523,
            'utt_idx': 'utt-1'
        }
        assert (tmp_path / all_content[2]['files'][0]['fname']).exists()

        train_content = json.loads(train_file.read_text())
        assert len(train_content) == 3

        dev_content = json.loads(dev_file.read_text())
        assert len(dev_content) == 1



import os

from audiomate.corpus.io import base

import pytest


def create_mock_corpus_downloader():

    class MockCorpusDownloader(base.CorpusDownloader):

        @classmethod
        def type(cls):
            return 'mock'

        def _download(self, target_path):
            os.makedirs(os.path.join(target_path, 'subfolder', 'a.txt'))

    return MockCorpusDownloader()


class TestCorpusDownloader:

    def test_force_redownload_overwrites_existing_directory(self, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = create_mock_corpus_downloader()

        tmpdir.mkdir('subfolder').join('b.txt')
        corpus_dl.download(target_folder, force_redownload=True)

        assert len(os.listdir(tmpdir)) == 1
        assert os.path.exists(os.path.join(target_folder, 'subfolder', 'a.txt'))
        assert not os.path.exists(os.path.join(target_folder, 'subfolder', 'b.txt'))

    def test_existing_directory_forces_io_error(self, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = create_mock_corpus_downloader()

        tmpdir.mkdir('subfolder').join('a.txt')

        with pytest.raises(IOError):
            corpus_dl.download(target_folder, force_redownload=False)


def create_mock_corpus_reader():

    class MockCorpusReader(base.CorpusReader):

        @classmethod
        def type(cls):
            return 'mock'

        def _load(self, path):
            return

        def _check_for_missing_files(self, path):
            return []

    return MockCorpusReader()


class TestCorpusReader:

    def test_invalid_path_forces_io_error(self, tmpdir):
        target_folder = tmpdir.strpath
        corpus_reader = create_mock_corpus_reader()

        with pytest.raises(IOError):
            path = os.path.join(target_folder, 'tmp')
            corpus_reader.load(path)



import abc
import copy
import os

import audiomate
from audiomate import tracks
from audiomate import logutil

logger = logutil.getLogger()


class AudioFileConverter(metaclass=abc.ABCMeta):
    """
    Base class for converters that convert all audio to a specific format.
    A converter creates a new instance of a corpus,
    so that all audio files meet given requirements.

    Args:
        sampling_rate (int): Target sampling rate to convert audio to.
        separate_file_per_utterance (bool): If ``True``, every utterance in the
                                       resulting corpus is in a separate file.
                                       If ``False``, the file/utt structure will
                                       be preserved.
        force_conversion (bool): If ``True``, all utterances will be converted
                                 whether or not it already matches the target
                                 format. If ``False``, only utterances not
                                 matching the target format will be converted.
                                 Others are reference to the original files.
    """

    def __init__(self, sampling_rate=16000, separate_file_per_utterance=False,
                 force_conversion=False):
        self.sampling_rate = sampling_rate
        self.separate_file_per_utterance = separate_file_per_utterance
        self.force_conversion = force_conversion

    def convert(self, corpus, target_audio_path):
        """
        Convert the given corpus.

        Args:
            corpus (Corpus): The input corpus.
            target_audio_path (str): The path where the audio files of the
                                     converted corpus should be saved.

        Returns:
            Corpus: The newly created corpus.
        """

        out_corpus = audiomate.Corpus()
        files_to_convert = []

        for utterance in logger.progress(
                corpus.utterances.values(),
                total=corpus.num_utterances,
                description='Find utterances to convert'):

            if utterance.issuer.idx not in out_corpus.issuers.keys():
                out_corpus.import_issuers(utterance.issuer)

            if self._does_utt_need_conversion(utterance):
                # Store audio in a new file

                if self.separate_file_per_utterance:
                    filename = '{}.{}'.format(utterance.idx, self._file_extension())
                    path = os.path.join(target_audio_path, filename)
                    files_to_convert.append((
                        utterance.track.path,
                        utterance.start,
                        utterance.end,
                        path
                    ))

                    track = out_corpus.new_file(path, utterance.idx)
                    start = 0
                    end = float('inf')

                else:
                    if utterance.track.idx not in out_corpus.tracks.keys():
                        filename = '{}.{}'.format(utterance.track.idx, self._file_extension())
                        path = os.path.join(target_audio_path, filename)
                        files_to_convert.append((
                            utterance.track.path,
                            0,
                            float('inf'),
                            path
                        ))
                        out_corpus.new_file(path, utterance.track.idx)

                    track = utterance.track
                    start = utterance.start
                    end = utterance.end

                utt = out_corpus.new_utterance(
                    utterance.idx,
                    track.idx,
                    issuer_idx=utterance.issuer.idx,
                    start=start,
                    end=end
                )

                lls = copy.deepcopy(list(utterance.label_lists.values()))
                utt.set_label_list(lls)

            else:
                # Just copy everything to the output corpus
                self._copy_utterance_to_corpus(utterance, out_corpus)

        self._copy_subviews_to_corpus(corpus, out_corpus)
        self._convert_files(files_to_convert)

        return out_corpus

    @abc.abstractmethod
    def _file_extension(self):
        """ Return the file-extension that will be used. """
        raise NotImplementedError()

    @abc.abstractmethod
    def _does_utt_match_target_format(self, utterance):
        """
        Return ``True`` if the utterance already matches the target format,
        ``False`` otherwise.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _convert_files(self, files):
        """
        Store the given samples with the target format
        at ``path``.
        """
        raise NotImplementedError()

    def _does_utt_need_conversion(self, utterance):
        """ Return True if an utterance needs to be converted. """
        if self.force_conversion:
            return True

        elif type(utterance.track) != tracks.FileTrack:
            return True

        elif self.separate_file_per_utterance and (utterance.start > 0 or utterance.end != float('inf')):
            return True

        elif not self._does_utt_match_target_format(utterance):
            return True

        return False

    def _copy_utterance_to_corpus(self, utterance, corpus):
        """ Create a copy of the utterance and add it to the given corpus. """

        if utterance.track.idx not in corpus.tracks.keys():
            corpus.import_tracks(utterance.track)

        corpus.import_utterances(utterance)

    def _copy_subviews_to_corpus(self, from_corpus, to_corpus):
        """ Create copy of all subviews from ``from_corpus`` in ``to_corpus``. """

        subviews = copy.deepcopy(from_corpus.subviews)
        for subview_idx, subview in subviews.items():
            to_corpus.import_subview(subview_idx, subview)



import functools
import multiprocessing

import soundfile as sf
import sox

from . import base
from audiomate import logutil

logger = logutil.getLogger()


class WavAudioFileConverter(base.AudioFileConverter):
    """
    Class that creates a new instance of a corpus,
    so that all audio files meet given requirements.
    """

    def __init__(self, num_workers=4, sampling_rate=16000, separate_file_per_utterance=False,
                 force_conversion=False):
        super(WavAudioFileConverter, self).__init__(
            sampling_rate,
            separate_file_per_utterance,
            force_conversion
        )

        self.num_workers = num_workers

        self.expected_properties = {
            'samplerate': self.sampling_rate,
            'format': 'WAV',
            'subtype': 'PCM_16'
        }

    def _file_extension(self):
        return 'wav'

    def _does_utt_match_target_format(self, utterance):
        """
        Return ``True`` if the utterance already matches the target format,
        ``False`` otherwise.
        """
        if utterance.track.path.endswith('mp3'):
            return False

        try:
            info = sf.info(utterance.track.path)

            for key, value in self.expected_properties.items():
                if info.__getattribute__(key) != value:
                    return False

        except RuntimeError:
            return False

        return True

    def _convert_files(self, files):
        """
        Store the given samples with the target format
        at ``path``.
        """
        with multiprocessing.Pool(self.num_workers) as p:
            func = functools.partial(
                _process_file,
                target_sr=self.sampling_rate
            )
            list(logger.progress(
                p.imap(func, list(files)),
                total=len(files),
                description='Convert audio files'
            ))


def _process_file(file_item, target_sr):
    src = file_item[0]
    start = file_item[1]
    end = file_item[2]
    target = file_item[3]

    tfm = sox.Transformer()

    if start > 0 and end == float('inf'):
        tfm.trim(start)
    elif end != float('inf'):
        tfm.trim(start, end)

    tfm.convert(target_sr, 1, 16)
    try:
        tfm.build(src, target)
    except sox.core.SoxError:
        logger.error('The following file could not be converted: %s', src)



import collections
import os

import audiomate
from audiomate.utils import textfile
from . import base
from audiomate.corpus import conversion


class Wav2LetterWriter(base.CorpusWriter):
    """
    Writes files to use for training/testing/decoding with
    wav2letter (https://github.com/facebookresearch/wav2letter).

    Since it is expected that every utterance is in a separate file,
    any utterances that are not in separate file in the original corpus,
    are extracted into a separate file in the
    subfolder `audio` of the target path.

    Args:
        export_all_audio (bool): If ``True``, all utterances are exported,
                                 whether they are in a separate file
                                 already or not.
        transcription_label_list_idx (str): The transcriptions are used from the label-list with this id.
        num_workers (int): Number of processes to use to process utterances.
    """

    def __init__(self, export_all_audio=False,
                 transcription_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT,
                 num_workers=1):
        self.export_all_audio = export_all_audio
        self.transcription_label_list_idx = transcription_label_list_idx
        self.sampling_rate = 16000
        self.num_workers = num_workers

        self.converter = conversion.WavAudioFileConverter(
            self.num_workers,
            self.sampling_rate,
            separate_file_per_utterance=True,
            force_conversion=self.export_all_audio
        )

    @classmethod
    def type(cls):
        return 'wav2letter'

    def _save(self, corpus, path):
        target_audio_path = os.path.join(path, 'audio')
        os.makedirs(target_audio_path, exist_ok=True)

        # Convert all files
        corpus = self.converter.convert(corpus, target_audio_path)
        records = []

        subset_utterance_ids = {idx: list(subset.utterances.keys()) for idx, subset in corpus.subviews.items()}
        subset_records = collections.defaultdict(list)

        for utterance_idx in sorted(corpus.utterances.keys()):
            utterance = corpus.utterances[utterance_idx]
            transcript = utterance.label_lists[self.transcription_label_list_idx].join()
            audio_path = utterance.track.path
            num_samples = int(utterance.duration * self.sampling_rate)

            # Add to the full list
            record = [utterance_idx, audio_path, num_samples, transcript]
            records.append(record)

            # Check / Add to subview lists
            for subset_idx, utt_ids in subset_utterance_ids.items():
                if utterance_idx in utt_ids:
                    subset_records[subset_idx].append(record)

        # Write full list
        records_path = os.path.join(path, 'all.lst')
        textfile.write_separated_lines(records_path, records, separator=' ', sort_by_column=-1)

        # Write subset lists
        for subset_idx, records in subset_records.items():
            if len(records) > 0:
                subset_file_path = os.path.join(path, '{}.lst'.format(subset_idx))
                textfile.write_separated_lines(subset_file_path, records, separator=' ', sort_by_column=-1)



import os

import pytest

import audiomate
from audiomate.corpus import io
from audiomate import issuers
from tests import resources


@pytest.fixture
def reader():
    return io.SWCReader()


@pytest.fixture
def sample_path():
    return resources.sample_corpus_path('swc')


class TestSWCReader:

    def test_get_articles(self, reader, sample_path):
        articles = reader.get_articles(sample_path)

        article_names = {
            'Fonteius_Capito',
            'Kritischer_Rationalismus',
            'Philosophie',
            'Radio',
        }

        article_paths = {os.path.join(sample_path, a) for a in article_names}

        assert set(articles) == article_paths

    def test_get_audio_file_info(self, reader, sample_path):
        article_path = os.path.join(sample_path, 'Kritischer_Rationalismus')
        audio_files = reader.get_audio_file_info(article_path)

        assert audio_files == {
            os.path.join(article_path, 'audio1.ogg'): 0.0,
            os.path.join(article_path, 'audio2.ogg'): 1224.0,
            os.path.join(article_path, 'audio3.ogg'): 2713.114375,
        }

    def test_get_reader_info_male(self, reader, sample_path):
        article_path = os.path.join(sample_path, 'Kritischer_Rationalismus')
        name, gender = reader.get_reader_info(article_path)

        assert name == 'Jan Krüger'
        assert gender == issuers.Gender.MALE

    def test_get_reader_info_female(self, reader, sample_path):
        article_path = os.path.join(sample_path, 'Fonteius_Capito')
        name, gender = reader.get_reader_info(article_path)

        assert name == 'Souffleuse'
        assert gender == issuers.Gender.FEMALE

    def test_find_audio_file_for_segment_if_only_one_audio_file(self, reader):
        audio_files = {
            'a': 0
        }
        res = reader.find_audio_file_for_segment(291, 302, audio_files)

        assert res == 'a'

    def test_find_audio_file_for_segment_if_in_first_audio_file(self, reader):
        audio_files = {
            'a': 0,
            'b': 193,
            'c': 340,
        }
        res = reader.find_audio_file_for_segment(89, 193, audio_files)

        assert res == 'a'

    def test_find_audio_file_for_segment_if_in_intermediate_audio_file(self, reader):
        audio_files = {
            'a': 0,
            'b': 291,
            'c': 340,
        }
        res = reader.find_audio_file_for_segment(291, 302, audio_files)

        assert res == 'b'

    def test_find_audio_file_for_segment_if_in_last_audio_file(self, reader):
        audio_files = {
            'a': 0,
            'b': 221,
            'c': 291,
        }
        res = reader.find_audio_file_for_segment(291, 302, audio_files)

        assert res == 'c'

    def test_find_audio_file_for_segment_if_across_audio_files(self, reader):
        audio_files = {
            'a': 0,
            'b': 221,
            'c': 291,
        }
        res = reader.find_audio_file_for_segment(279, 302, audio_files)

        assert res is None

    def test_get_segments(self, reader, sample_path):
        article_path = os.path.join(sample_path, 'Fonteius_Capito')
        segments = reader.get_segments(article_path)

        exp = [
            (1.010, 3.620, 'Sie hören den Artikel Fonteius Capito'),
            (27.780, 30.680, 'Im Jahr siebenundsechzig nach Christus'),
            (73.530, 78.340, 'dem Tod Neros wurde Fonteius Capito wegen angeblicher Umsturzpläne'),
        ]

        assert sorted(segments, key=lambda x: x[0]) == exp

    def test_load(self, reader, sample_path):
        ds = reader.load(sample_path)

        assert ds.num_issuers == 4
        assert ds.num_utterances == 2055
        assert ds.num_tracks == 7

        utt = None

        for utterance in ds.utterances.values():
            if utterance.idx.endswith('4338630_4340650'):
                utt = utterance

        assert utt.start == 4338.63 - 2713.114375
        assert utt.end == 4340.650 - 2713.114375

        ll = utt.label_lists[audiomate.corpus.LL_WORD_TRANSCRIPT]

        assert len(ll) == 1
        assert ll.join() == 'Bartley vertrat die Auffassung'



import multiprocessing

import audioread

from . import base
from audiomate import logutil

logger = logutil.getLogger()


class TrackReadValidator(base.Validator):
    """
    Check if the track can be opened and read.
    By reading the first few samples.
    """

    def __init__(self, num_workers=1):
        self.num_workers = num_workers

    def name(self):
        return 'Track-Read'

    def validate(self, corpus_to_validate):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            InvalidItemsResult: Validation result.
        """

        with multiprocessing.pool.ThreadPool(self.num_workers) as p:
            result = list(logger.progress(
                p.imap(
                    self.validate_track,
                    list(corpus_to_validate.tracks.values())
                ),
                total=corpus_to_validate.num_tracks,
                description='Validate tracks'
            ))

        invalid_tracks = {x[0]: x[1] for x in result if x[1] is not None}
        passed = len(invalid_tracks) <= 0

        return base.InvalidItemsResult(
            passed,
            invalid_tracks,
            item_name='Tracks',
            name=self.name()
        )

    def validate_track(self, track):
        result = None

        try:
            track.read_samples(duration=0.001)
        except EOFError:
            result = 'EOFError'
        except audioread.NoBackendError:
            result = 'NoBackendError'

        # skipcq: PYL-W0703
        # Basically we want to get all causes of a file not being read.
        # We should try to figure out all possible ones.
        except Exception as ex:
            result = str(ex)

        return (track.idx, result)



import librosa
import numpy as np

from tests import resources

from audiomate import tracks
from audiomate.corpus import conversion


class TestWavAudioFileConverter:

    def test_does_utt_match_target_format_returns_true(self):
        file_path = resources.sample_wav_file('wav_1.wav')
        track = tracks.FileTrack('t', file_path)
        utt = tracks.Utterance('u', track)

        c = conversion.WavAudioFileConverter()
        assert c._does_utt_match_target_format(utt)

    def test_does_utt_match_target_format_with_invalid_sr_returns_false(self):
        file_path = resources.get_resource_path(('audio_formats', 'wav_2_44_1k_16b.wav'))
        track = tracks.FileTrack('t', file_path)
        utt = tracks.Utterance('u', track)

        c = conversion.WavAudioFileConverter()
        assert not c._does_utt_match_target_format(utt)

    def test_does_utt_match_target_format_with_invalid_format_returns_false(self):
        file_path = resources.get_resource_path(('audio_formats', 'mp3_2_44_1k_16b.mp3'))
        track = tracks.FileTrack('t', file_path)
        utt = tracks.Utterance('u', track)

        c = conversion.WavAudioFileConverter()
        assert not c._does_utt_match_target_format(utt)

    def test_convert_files(self, tmp_path):
        source_path = resources.sample_wav_file('wav_1.wav')
        target_path = tmp_path / 'out.wav'

        files = [(source_path, 0, float('inf'), str(target_path))]

        c = conversion.WavAudioFileConverter()
        c._convert_files(files)

        samples, sr = librosa.core.load(source_path, sr=None)

        stored_samples, stored_sr = librosa.core.load(str(target_path), sr=None)

        assert target_path.is_file()
        assert stored_sr == sr
        assert np.array_equal(stored_samples, samples)

    def test_store_samples_sr_24(self, tmp_path):
        source_path = resources.sample_wav_file('wav_1.wav')
        target_path = tmp_path / 'out.wav'

        files = [(source_path, 0, float('inf'), str(target_path))]

        c = conversion.WavAudioFileConverter(sampling_rate=24000)
        c._convert_files(files)

        samples, sr = librosa.core.load(source_path, sr=24000)

        stored_samples, stored_sr = librosa.core.load(str(target_path), sr=None)

        assert target_path.is_file()
        assert stored_sr == sr
        # Don't take it too exactly
        # With sox 14.4.1 it isn't that precise, expecially the first sample
        assert np.allclose(stored_samples[1:], samples[1:], atol=0.001)



"""
This module contains any functionality for downloading and
extracting data from any remotes.
"""

import zipfile
import tarfile
import multiprocessing

import requests
from pget.down import Downloader

from audiomate import logutil

logger = logutil.getLogger()

PROGRESS_LOGGER_BYTE_DELAY = 1024 * 1024 * 100


def download_files(url_to_target, num_threads=1):
    """
    Download multiple files.

    Args:
        url_to_target (dict): Dict with mapping from source-url
                              to target-path.
        num_threads (int): Number of threads to use.
    """
    dl_items = list(url_to_target.items())

    with multiprocessing.pool.ThreadPool(num_threads) as p:
        result = list(logger.progress(
            p.imap(_download_file, dl_items),
            total=len(dl_items),
            description='Download Files'
        ))

        return result


def _download_file(item):
    """ Helper function to pass (url, target) to ``download_file``. """
    return download_file(item[0], item[1])


def download_file(url, target_path, num_threads=1):
    """
    Download the file from the given `url` and store it at `target_path`.
    Return a tuple x (url, bool, str).
    x[0] contains the url.
    If download failed x[1] is ``False`` and x[2] contains some error message.
    If download was fine x[1] is ``True`` and x[2] contains the target-path.
    """

    if num_threads > 1:
        return download_file_parallel(
            url,
            target_path,
            num_threads=num_threads
        )

    r = requests.get(url, stream=True)

    if r.status_code != 200:
        return (url, False, 'Failed to download file {} (status {})!'.format(
            r.status_code,
            url
        ))

    file_size = int(requests.head(url).headers['Content-Length'])
    bytes_loaded = 0
    bytes_since_last_log = 0
    logger.info('Download file from "%s" with size: %d B', url, file_size)

    with open(target_path, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                bytes_loaded += 1024
                bytes_since_last_log += 1024

                if bytes_since_last_log >= PROGRESS_LOGGER_BYTE_DELAY:
                    logger.info('Download [%06.2f%%]', bytes_loaded / file_size * 100)
                    bytes_since_last_log = 0

    logger.info('Finished download')

    return (url, True, target_path)


def download_file_parallel(url, target_path, num_threads=1):
    """
    Download the file from the given `url` and store it at `target_path`.
    Return a tuple x (url, bool, str).
    x[0] contains the url.
    If download failed x[1] is ``False`` and x[2] contains some error message.
    If download was fine x[1] is ``True`` and x[2] contains the target-path.
    """

    downloader = Downloader(url, target_path, num_threads)
    downloader.start()

    # Wait until we know file size
    while downloader.total_length == 0:
        pass

    file_size = downloader.total_length
    logger.info('Download file from "%s" with size: %d B', url, file_size)

    bytes_at_last_log = 0

    def callback(x):
        nonlocal bytes_at_last_log

        if x.total_downloaded - bytes_at_last_log >= PROGRESS_LOGGER_BYTE_DELAY:
            logger.info('Download [%06.2f%%]', x.total_downloaded / file_size * 100)
            bytes_at_last_log = x.total_downloaded

    downloader.subscribe(callback, 10)
    downloader.wait_for_finish()

    logger.info('Finished download')

    return (url, True, target_path)


def extract_zip(zip_path, target_folder):
    """
    Extract the content of the zip-file at ``zip_path`` into
    ``target_folder``.
    """
    with zipfile.ZipFile(zip_path) as archive:
        archive.extractall(target_folder)


def extract_tar(tar_path, target_folder):
    """
    Extract the content of the tar-file at ``tar_path`` into
    ``target_folder``.
    """
    with tarfile.open(tar_path, 'r') as archive:
        archive.extractall(target_folder)



import os
import json
import multiprocessing

import librosa

import audiomate
from . import base

from audiomate.corpus import conversion
from audiomate import logutil

logger = logutil.getLogger()


class NvidiaJasperWriter(base.CorpusWriter):
    """
    Writes files to use for training with NVIDIA Jasper
    (https://github.com/NVIDIA/DeepLearningExamples/tree/master/PyTorch/SpeechRecognition/Jasper).

    **Audio Files:**
    Since utterances are expected to be in a separate file and have
    a specific format (WAV 16-Bit PCM), any utterances that do not meet those
    requirements, are extracted into a separate file in the
    subfolder `audio` of the target path.

    **Subviews:**
    Subviews in the Jasper format are represented as different
    Json-Files. For every subview in the corpus a separate
    Json-File is created. Additionaly, a ``all.json`` is created,
    that contains all utterances.

    Args:
        data_base_path (str): Path from where the audio files are found.
                              If ``None`` it is the path where the corpus is saved.
        no_audio_check (bool): If ``True``, the audio is not check for correct format.
        export_all_audio (bool): If ``True``, all utterances are exported,
                                 whether they are in a separate file
                                 already or not.
        transcription_label_list_idx (str): The transcriptions are used from
                                            the label-list with this id.
        sampling_rate (int): Target sampling rate to use.
        num_workers (int): Number of processes to use to process utterances.
    """

    def __init__(self, data_base_path=None,
                 no_check=False,
                 export_all_audio=False,
                 transcription_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT,
                 sampling_rate=16000,
                 num_workers=1):

        self.data_base_path = data_base_path
        self.no_check = no_check
        self.export_all_audio = export_all_audio
        self.transcription_label_list_idx = transcription_label_list_idx
        self.sampling_rate = sampling_rate
        self.num_workers = num_workers

        self.converter = conversion.WavAudioFileConverter(
            self.num_workers,
            self.sampling_rate,
            separate_file_per_utterance=True,
            force_conversion=self.export_all_audio
        )

    @classmethod
    def type(cls):
        return 'nvidia-jasper'

    def _save(self, corpus, path):
        target_audio_path = os.path.join(path, 'audio')
        os.makedirs(target_audio_path, exist_ok=True)

        if self.no_check:
            out_corpus = corpus
        else:
            out_corpus = self.converter.convert(corpus, target_audio_path)

        if self.data_base_path:
            rel_base_path = self.data_base_path
        else:
            rel_base_path = path

        utts = []

        logger.info('Get utterance durations')
        utt_idx_to_path = {utt.idx: utt.track.path for utt in out_corpus.utterances.values()}
        utt_to_duration = self._get_file_durations(utt_idx_to_path)

        logger.info('Prepare utterance infos')
        for utterance in out_corpus.utterances.values():
            utt_dur = utt_to_duration[utterance.idx]
            utt_info = self.process_utterance(utterance, utt_dur, rel_base_path)
            utts.append((utterance.idx, utt_info))

        utts = sorted(utts, key=lambda x: x[1]['original_duration'])

        logger.info('Prepare subviews')
        subviews = {}
        subviews['all'] = [u[1] for u in utts]

        for subview_name, subview in corpus.subviews.items():
            utt_filter = subview.utterances.keys()
            subview_utts = [u[1] for u in utts if u[0] in utt_filter]
            subviews[subview_name] = subview_utts

        logger.info('Write files')
        for name, data in subviews.items():
            target_path = os.path.join(path, '{}.json'.format(name))
            with open(target_path, 'w') as f:
                json.dump(data, f)

    def process_utterance(self, utt, utt_dur, rel_base_path):
        rel_path = os.path.relpath(utt.track.path, rel_base_path)
        num_samples = utt_dur * self.sampling_rate

        return {
            'transcript': utt.label_lists[self.transcription_label_list_idx].join(),
            'files': [{
                'fname': rel_path,
                'channels': 1,
                'sample_rate': 16000,
                'duration': utt_dur,
                'num_samples': num_samples,
                'speed': 1
            }],
            'original_duration': utt_dur,
            'original_num_samples': num_samples,
            'utt_idx': utt.idx
        }

    def _get_file_durations(self, files):
        file_items = files.items()
        num_files = len(file_items)

        with multiprocessing.Pool(self.num_workers) as p:
            res = list(logger.progress(
                p.imap(self._get_file_duration, file_items),
                total=num_files
            ))

        return {k: v for (k, v) in res}

    def _get_file_duration(self, file_item):
        dur = librosa.core.get_duration(filename=file_item[1])
        return (file_item[0], dur)



"""
This module contains classes to convert the data of a corpus.
It is for example used to convert all audio data to wav files.

Audio File Conversion
---------------------
.. autoclass:: AudioFileConverter
   :members:
   :inherited-members:

.. autoclass:: WavAudioFileConverter
   :members:
   :inherited-members:

"""

from .base import AudioFileConverter  # noqa: F401

from .wav import WavAudioFileConverter  # noqa: F401



import collections
import os
import glob
import re

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus.subset import subview
from . import base
from . import downloader

DOWNLOAD_URL = 'http://ltdata1.informatik.uni-hamburg.de/kaldi_tuda_de/german-speechdata-package-v3.tar.gz'


SPEAKER_IDX_PATTERN = re.compile(r'<speaker_id>(.*?)</speaker_id>')
GENDER_PATTERN = re.compile(r'<gender>(.*?)</gender>')
TRANSCRIPTION_PATTERN = re.compile(r'<cleaned_sentence>(.*?)</cleaned_sentence>')
RAW_TRANSCRIPTION_PATTERN = re.compile(r'<sentence>(.*?)</sentence>')
AGE_PATTERN = re.compile(r'<ageclass>(.*?)</ageclass>')
NATIVE_PATTERN = re.compile(r'<muttersprachler>(.*?)</muttersprachler>')

SUBSETS = ['train', 'dev', 'test']

WAV_FILE_SUFFIXES = [
    'Kinect-Beam',
    'Kinect-RAW',
    'Realtek',
    'Samson',
    'Yamaha',
    'Microsoft-Kinect-Raw'
]

WAV_SUFFIX_TO_SUBVIEW = {
    'Kinect-Beam': 'kinect-beam',
    'Kinect-RAW': 'kinect-raw',
    'Realtek': 'realtek',
    'Samson': 'samson',
    'Yamaha': 'yamaha',
    'Microsoft-Kinect-Raw': 'kinect-raw',
}


class TudaDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the TUDA Corpus.

    Args:
        url (str): The url to download the dataset from. If not given the
                   default URL is used. It is expected to be a tar.gz file.
        num_threads (int): Number of threads to use for download files.
    """

    def __init__(self, url=None, num_threads=1):
        if url is None:
            url = DOWNLOAD_URL

        super(TudaDownloader, self).__init__(
            url,
            move_files_up=True,
            num_threads=num_threads
        )

    @classmethod
    def type(cls):
        return 'tuda'


class TudaReader(base.CorpusReader):
    """
    Reader for the TUDA german distant speech corpus.

    Note:
        It only loads files ending in -beamformedSignal.wav

    .. seealso::

       `<https://www.inf.uni-hamburg.de/en/inst/ab/lt/resources/data/acoustic-models.html>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'tuda'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for part in SUBSETS:
            sub_path = os.path.join(path, part)
            ids = self.get_ids_from_folder(sub_path)
            utt_ids = []

            for idx in ids:
                add_ids = self.load_file(sub_path, idx, corpus)
                utt_ids.extend(add_ids)

            subview_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=utt_ids)
            subview_corpus = subview.Subview(corpus, filter_criteria=[subview_filter])
            corpus.import_subview(part, subview_corpus)

            TudaReader.create_wav_type_subviews(corpus, utt_ids, prefix='{}_'.format(part))

        TudaReader.create_wav_type_subviews(corpus, corpus.utterances.keys())

        return corpus

    @staticmethod
    def create_wav_type_subviews(corpus, utt_ids, prefix=''):
        splits = collections.defaultdict(list)

        for utt_id in utt_ids:
            wavtype = utt_id.split('_')[-1]
            subview_type = WAV_SUFFIX_TO_SUBVIEW[wavtype]
            splits[subview_type].append(utt_id)

        for sub_name, sub_utts in splits.items():
            subview_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=sub_utts)
            subview_corpus = subview.Subview(corpus, filter_criteria=[subview_filter])
            corpus.import_subview('{}{}'.format(prefix, sub_name), subview_corpus)

    def get_ids_from_folder(self, path):
        """
        Return all ids from the given folder,
        which have a corresponding beamformedSignal file.
        """
        ids = set()

        for xml_file in glob.glob(os.path.join(path, '*.xml')):
            idx = os.path.splitext(os.path.basename(xml_file))[0]
            ids.add(idx)

        return ids

    def load_file(self, folder_path, idx, corpus):
        """
        Load speaker, file, utterance, labels
        for the file with the given id.
        """
        xml_path = os.path.join(folder_path, '{}.xml'.format(idx))
        wav_paths = []

        for wav_suffix in WAV_FILE_SUFFIXES:
            wav_path = os.path.join(folder_path, '{}_{}.wav'.format(idx, wav_suffix))
            wav_name = os.path.split(wav_path)[1]
            wav_idx = os.path.splitext(wav_name)[0]

            if os.path.isfile(wav_path) and wav_idx not in self.invalid_utterance_ids:
                wav_paths.append(wav_path)

        if len(wav_paths) == 0:
            return []

        with open(xml_path, 'r', encoding='utf-8') as f:
            text = f.read()

        transcription = TudaReader.extract_value(text, TRANSCRIPTION_PATTERN, 'transcription', xml_path)
        transcription_raw = TudaReader.extract_value(text, RAW_TRANSCRIPTION_PATTERN, 'raw_transcription', xml_path)
        gender = TudaReader.extract_value(text, GENDER_PATTERN, 'gender', xml_path)
        is_native = TudaReader.extract_value(text, NATIVE_PATTERN, 'native', xml_path)
        age_class = TudaReader.extract_value(text, AGE_PATTERN, 'age', xml_path)
        speaker_idx = TudaReader.extract_value(text, SPEAKER_IDX_PATTERN, 'speaker_idx', xml_path)

        if speaker_idx not in corpus.issuers.keys():
            start_age_class = int(age_class.split('-')[0])

            if start_age_class < 12:
                age_group = issuers.AgeGroup.CHILD
            elif start_age_class < 18:
                age_group = issuers.AgeGroup.YOUTH
            elif start_age_class < 65:
                age_group = issuers.AgeGroup.ADULT
            else:
                age_group = issuers.AgeGroup.SENIOR

            native_lang = None

            if is_native == 'Ja':
                native_lang = 'deu'

            issuer = issuers.Speaker(speaker_idx,
                                     gender=issuers.Gender(gender),
                                     age_group=age_group,
                                     native_language=native_lang)
            corpus.import_issuers(issuer)

        utt_ids = []

        for wav_path in wav_paths:
            wav_name = os.path.split(wav_path)[1]
            wav_idx = os.path.splitext(wav_name)[0]
            corpus.new_file(wav_path, wav_idx)
            utt = corpus.new_utterance(wav_idx, wav_idx, speaker_idx)
            utt.set_label_list(annotations.LabelList.create_single(
                transcription,
                idx=audiomate.corpus.LL_WORD_TRANSCRIPT
            ))
            utt.set_label_list(annotations.LabelList.create_single(
                transcription_raw,
                idx=audiomate.corpus.LL_WORD_TRANSCRIPT_RAW
            ))
            utt_ids.append(wav_idx)

        return utt_ids

    @staticmethod
    def extract_value(text, pattern, value, path):
        m = pattern.search(text)

        if m:
            return m.group(1)
        else:
            raise ValueError('Value {} not found in {}'.format(value, path))
